/* vi: expandtab:softtabstop=4:tabstop=4:cindent:cino=>4,g4,h4,+4,c1,C4,(4,u4,{0,^-0
 */

/*

Copyright (c) 2008, SFB/TR-55
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef FEENLIB_H_
#define FEENLIB_H_

/*
 * Private Note:
 * Why is this thingy called "feenlib" ?
 * It's derived from "Front End ENchanted LIBrary
 */

/** @file feenlib.h
 * This file contains all definitions necessary to
 * interact with the @b feenlib library.
 *
 * You will notice that many libary calls take a parameter
 * of the form
 *
 * <code>uint8_t * data</code>
 *
 * Although you can transmit an arbitrary amount of data that way,
 * the library will operate differently internally, depending on
 * how many bytes are actually transferred.
 * Typcially, transferring less than 1000 bytes is fast,
 * but if one transfers more than 1000 bytes, the data needs to be
 * buffered on the client (rootcard) side.
 */

#define FEEN_SVN_ID_OF_FEENLIB_H "$Id$"

/** Definition for SPI handling.
 * See @a feenSPIShift for details.
 */
#define FEEN_CS_AUTO_OFF 0xCCC2

/** Definition for SPI handling.
 * See @a feenSPIShift for details.
 */
#define FEEN_CS_AUTO_1   0xCCC1

/** Definition for SPI handling.
 * See @a feenSPIShift for details.
 */
#define FEEN_CS_AUTO_0   0xCCC0

/** Definition for UART handling.
 * See @a feenUARTWrite for details.
 */
#define FEEN_UART_RAW   0xAAA0

/** Definition for UART handling.
 * See @a feenUARTWrite for details.
 */
#define FEEN_UART_TEXT  0xAAA1

/** Definition for UART handling.
 * See @a feenUARTWrite for details.
 */
#define FEEN_UART_EVEN  0xAAA2

/** Definition for UART handling.
 * See @a feenUARTWrite for details.
 */
#define FEEN_UART_ODD   0xAAA3

/** Definition for UART handling.
 * See @a feenUARTWrite for details.
 */
#define FEEN_UART_MARK  0xAAA4

/** Definition for UART handling.
 * See @a feenUARTWrite for details.
 */
#define FEEN_UART_SPACE 0xAAA5

/** Definition for UART handling.
 * See @a feenUARTWrite for details.
 */
#define FEEN_UART_NONE  0xAAA6

/** Rootcard scans for a new packet.
 * See @a feenStatus for details.
 */
#define FEEN_SCANNING   0x88000001 /* keep in sync with ROBPROT_SCANNING */

/** Rootcard processes internal command.
 * See @a feenStatus for details.
 */
#define FEEN_INTER_WIP  0x88000002 /* keep in sync with ROBPROT_INTER_WIP */

/** Rootcard processes external command.
 * See @a feenStatus for details.
 */
#define FEEN_EXTERN_WIP 0x88000003 /* keep in sync with ROBPROT_EXTERN_WIP */

/** Rootcard is idle.
 * See @a feenStatus for details.
 */
#define FEEN_IDLE       0x88000004 /* keep in sync with ROBPROT_IDLE */

/** Default IP port for data.
 * The underlying protocol uses this port by default
 * for transferring data between the rootcard and 
 * the frontend.
 */
#define FEEN_DEFAULT_DATAPORT     30601

/** Default IP port for contolwords.
 * The underlying protocol uses this port by default
 * for transferring control words between the rootcard
 * and the frontend.
 */
#define FEEN_DEFAULT_CONTROLPORT  30602

/** Maximum number of bytes for sequential transfers.
 *
 * This macro holds the maximum number of bytes that can
 * be transferred in sequential
 * transfers, such as SPI and IIC transfers.
 *
 * This is to be kept in sync with the definion of
 * "maxPayload" in Robprot.asn1
 */
#define FEEN_MAX_SEQUENTIAL_DATA 200000

/** Sector size of the internal flash.
 */
#define FEEN_SECTOR_SIZE_FLASH_INTERNAL  2048

/** Sector size of the exteranl flash.
 */
#define FEEN_SECTOR_SIZE_FLASH_EXTERNAL  65536

/** Flash pagesize.
 */
#define FEEN_PAGE_SIZE_FLASH 256

/** Protcol Version.
 */
#define FEEN_PROTO_VERSION 2

/* 
 * The public interface only makes sense on the frontend,
 * which is probably not running uClinux.
 * However, the macros and constants are shared between 
 * the rootcard and the frontend.
 */
#ifndef __uClinux__

#include <sys/socket.h>
#include <netinet/in.h>
#include <stdint.h>
#include <unistd.h>

/** Descriptor type for rootcards.
 * This type is used to identify a given rootcard.
 */
typedef struct feenRootCard {
    int timeout;    /**< timeout in seconds */
    int data_fd;    /**< file descriptor for data I/O */
    int control_fd; /**< file descriptor for control I/O */
    struct sockaddr data_sa; /**< socket structure for data I/O */
    struct sockaddr control_sa; /**< socket structore for control I/O */
} feenRC_t;


/** Get the numeric address of a rootcard.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [out] message See @a feenIICWrite.
 * @return 
 *      address of the rootcard, as returned by the POSIX function
 *      @a inet_aton. It is a statically allocated buffer, so the user
 *      must not "free" it, and should copy it for later use, because
 *      subsequent function calls will overwrite it.
 */
extern char* feenRootCardNumeric (
    feenRC_t rc,
    char ** const message
    );

/** Get the data port of a rootcard.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [out] message See @a feenIICWrite.
 * @return
 *      The data port of the rootcard in host byte order,
 *      or @c 0 in case of failure.
 */
extern uint16_t feenRootCardDataport (
    feenRC_t rc,
    char ** const message
    );

/** Get the control port of a rootcard.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [out] message See @a feenIICWrite.
 * @return
 *      The control port of the rootcard in host byte order,
 *      or @c 0 in case of failure.
 */
uint16_t feenRootCardControlport (
    feenRC_t rc,
    char ** const message
    );

/** IIC write command.
 *
 * @param [in] rc
 *      This parameter identifies the rootcard to talk to.
 *      It saves the sockets and additional parameters necessary
 *      to provide an efficient communication with the rootcard
 *      in question.
 * @param [in] addr
 *      IIC address to write to.
 *      Only the last 7 bits are taken into account.
 * @param [in] data
 *      bytes to be written.
 * @param length [in]
 *      number of bytes in @a data
 * @param [out] message
 *      Error Message. If this parameter is @a NULL,
 *      then no error message will be returned.
 *      Otherwise, the returned @a message parameter will point to a
 *      null-terminated string that can be used with @a printf.
 *      the user should not try to @a free the returned message.
 * @return
 *      The number of bytes that have been written,
 *      or @c -1 in case of failure.
 */
extern ssize_t feenIICWrite (
    feenRC_t rc,
    uint8_t addr,
    const uint8_t * data,
    size_t length,
    char ** const message
    );

/** IIC read command.
 *
 * @param [in] rc see @a feenIICWrite.
 * @param [in] addr see @a feenIICWrite.
 * @param [out] data buffer where the read data will be stored
 * @param [in] length number of bytes to be read
 * @param [out] message see @a feenIICWrite.
 * @return
 *      The number of bytes that have been read,
 *      or @c -1 in case of failure.
 */
 extern ssize_t feenIICRead  (
     feenRC_t rc,
     uint8_t addr,
     uint8_t * data,
     size_t length,
     char ** const message
     );

/** Low-level SPI command.
 *
 * @param [in] rc see @a feenIICWrite.
 * @param [in] chipselect
 *      This parameter determines how the chipselect lines
 *      are handled. A parameter of
 *      @a FEEN_CS_AUTO_OFF means, that one has to handle the
 *      chipselect lines manually, using
 *      @a feenSPIChip.
 *      @a FEEN_CS_AUTO_0 means, that @a feenSPIShift
 *      handles /CS0 automatically, while /CS1 is set
 *      to "high", i.e., /CS1 is inactive.
 *      @a FEEN_CS_AUTO_1 means, that @a feenSPIShift
 *      handles /CS1 automatically, while /CS0 is set
 *      to "high", i.e., /CS0 is inactive.
 * @param [in,out] data
 *      data to be shifted in.
 * @param [in] length
 *      length of @a data
 * @param [in] reply
 *      This parameter determines if the SPI reply,
 *      i.e., the data that is shifted out during
 *      the SPI process, is returned.
 *      If @a reply is logic true (1), then @a data
 *      is replaced by the SPI reply,
 *      if @a reply is logic false (0), then @a data
 *      is left untouched.
 * @param [out] message
 *      see @a feenIICWrite
 * @return
 *      The number of bytes that have been shifted
 *      in/out, or @c -1 in case of failure.
 */
extern ssize_t feenSPIShift (
    feenRC_t rc,
    int chipselect,
    uint8_t * data,
    size_t length,
    int reply,
    char ** const message
    );

/** SPI low level routine to access the /CS lines.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] chipselect
 *      What chipselect line to use.
 *      Valid values are @c 0 and @c 1.
 * @param [in] highlow
 *      What state the /CS line should be in.
 *      setting @a highlow to @c 0 activates the
 *      corresponding chip, since chipselect lines are
 *      active low.
 * @param [out] message See @a feenIICWrite.
 * @return @c 0 in case of success
 * @return @c -1 in case of failure.
 */
extern int feenSPIChip  (
    feenRC_t rc,
    int chipselect,
    int highlow,
    char ** const message
    );

/** Flash convenience routine for main SPI flash.
 *
 * This routine offers a convenience wrapper for
 * writing data to the SPI flash mounted on the nodecard.
 *
 * Note that @a offset has to be a multiple of
 * @a FEEN_SECTOR_SIZE_FLASH_EXTERNAL.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] offset
 *      Offset into the flash memory.
 *      Typically, the FPGA bitstream starts at offset 0x0,
 *      and the host firmware starts at offset 0x40000, but
 *      check the corresponding specs.
 * @param [in] data Data to be written.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return
 *      Number of bytes actually written,
 *      or @c -1 in case of failure.
 */
extern ssize_t feenFlashExternalWrite (
    feenRC_t rc,
    uint32_t offset,
    const uint8_t * data,
    size_t length,
    char ** const message
    );

/** Flash convenience routine with callbacks.
 *
 * This function is identical to @a feenFlashExternalWrite,
 * but provides callback function parameters, in order to provide
 * some status information. If @a sectrocallback and @a pagecallback
 * are passed a @a NULL, then this function is identical to 
 * @a feenFlashExternalWrite.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] offset
 *      Offset into the flash memory.
 *      Typically, the FPGA bitstream starts at offset 0x0,
 *      and the host firmware starts at offset 0x40000, but
 *      check the corresponding specs.
 * @param [in] data Data to be written.
 * @param [in] length Length of @a data.
 * @param [in] sectorcallback once a sector is written, this
 *      function will be invoked if it is not @a NULL. The sector
 *      number is passed as parameter.
 * @param [out] message See @a feenIICWrite.
 * @return
 *      Number of bytes actually written,
 *      or @c -1 in case of failure.
 */
extern ssize_t feenFlashExternalWriteCallbacks (
    feenRC_t rc,
    uint32_t offset,
    const uint8_t * data,
    size_t length,
    void (*sectorcallback) (int*),
    char ** const message
    );

/** Flash convenience routine for main SPI flash.
 *
 * This routine offers a convenience wrapper for
 * reading data from the SPI flash mounted on the nodecard.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] offset
 *      Offset into the flash memory.
 *      Typically, the FPGA bitstream starts at offset 0x0,
 *      and the host firmware starts at offset 0x40000, but
 *      check the corresponding specs.
 * @param [out] data Data to be written.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return
 *      Number of bytes actually read,
 *      or @c -1 in case of failure.
 */
extern ssize_t feenFlashExternalRead (
    feenRC_t rc,
    uint32_t offset,
    uint8_t * data,
    size_t length,
    char ** const message
    );

/** Flash convenience routine for internal SP Flash.
 *
 * This routine offers a convenice wrapper for 
 * writing data to the internal flash memory of 
 * the service processor. Please note that this routine 
 * always performs a <b>bulk erase</b> before each write.
 * @param [in] rc See @a feenIICWrite.
 * @param [in] offset Offset into the flash memory.
 * @param [in] data Data to be written.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return 
 *      Number of bytes actually written,
 *      or @c -1 in case of failure.
 */
extern ssize_t feenFlashInternalWrite (
    feenRC_t rc,
    uint32_t offset,
    const uint8_t * data,
    size_t length,
    char ** const message
    );

/** Flash convenience routine for internal SP Flash.
 *
 * This routine offers a convenice wrapper for 
 * reading data from the internal flash memory of 
 * the service processor. Please note that this routine 
 * always performs a <b>bulk erase</b> before each write.
 * @param [in] rc See @a feenIICWrite.
 * @param [in] offset Offset into the flash memory.
 * @param [out] data Data to be read.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return 
 *      Number of bytes actually read,
 *      or @c -1 in case of failure.
 */
extern ssize_t feenFlashInternalRead (
    feenRC_t rc,
    uint32_t offset,
    uint8_t * data,
    size_t length,
    char ** const message
    );

/** General UART write.
 * @deprecated Not yet implemented.
 * @param [in] rc See @a feenIICWrite.
 * @param [in] mode
 *      What mode to be used for accessing the
 *      UART. There are two possible modes:
 *      @a FEEN_UART_RAW and @a FEEN_UART_TEXT.
 *      @a FEEN_UART_RAW uses raw UART access on the 
 *      rootcard side, i.e., the UART is put in the mode
 *      as returned by the POSIX function @a cfmakeraw.
 *      @a FEEN_UART_TEXT tries to put the UART in a 
 *      sensible for text I/O.
 * @param [in] uart
 *      Which UART to use. Valid values are:
 *      -   @c 0 maps to /dev/ttyS0 on the rootcard, which is
 *          the console of the rootcard controller.
 *      -   @c 1 maps to /dev/ttyS1 on the rootcard, which is
 *          connected to the RS485-bus.
 *      -   @c 2 maps to /dev/ttyS2, which is connected to the 
 *          FPGA.
 * @param [in] speed
 *      Baud rate to use. Valid values are:<br>
 *      50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400,
 *      4800, 9600, 19200, 38400, 57600, 115200
 * @param [in] parity
 *      Parity to use. Valid values are:
 *      - @a FEEN_UART_EVEN  even parity
 *      - @a FEEN_UART_ODD   odd parity
 *      - @a FEEN_UART_MARK  mark parity (parity bit is always 1)
 *      - @a FEEN_UART_SPACE space parity (parity bit is always 0)
 *      - @a FEEN_UART_NONE  no parity
 * @param [in] data
 *      Data to be written.
 * @param [in] length
 *      Length of @a data.
 * @param [out] message
 *      See @a feenIICWrite.
 * @return Number of bytes actually written, or @c -1 in case of failure.
 */
extern ssize_t feenUARTWrite(
    feenRC_t rc,
    int mode,
    int uart,
    int speed,
    int parity,
    const uint8_t * data,
    size_t length,
    char ** const message
    );

/** General UART read.
 * @deprecated Not yet implemented.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] mode See @a feenUARTWrite.
 * @param [in] uart See @a feenUARTWrite.
 * @param [in] speed See @a feenUARTWrite.
 * @param [in] parity See @a feenUARTWrite.
 * @param [out] data Read data will be stored here.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return Number of bytes actually read, or @c -1 in case of failure.
 */
extern ssize_t feenUARTRead (
    feenRC_t rc,
    int mode,
    int uart,
    int speed,
    int parity,
    uint8_t * data,
    size_t length,
    char ** const message
    );

/** Convenience UART function.
 * @deprecated Not yet implemented.
 *
 * This is a convenice function talk to the service processor.
 * It saves the user to use @a feenUARTWrite function and 
 * takes care of the correct parity and speed settings.
 * However, bytes are sent directly to the service processor,
 * no package processing is performed, i.e., the user has to 
 * take care of correct SOF, EOF, STAT, etc. fields herself.
 *
 * If the user wishes a function that constructs RS485 packages
 * for the service processor, see @a feenServiceCommand.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] addr
 *      Address on the RS485 bus. Valid values range from 
 *      @c 0 to @c 15.
 *      Special exception: If @a addr is @c 0xFF, then no 
 *      address bytes are written.
 * @param [in] data Data to be written.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return Number of bytes actually written, or @c -1 in case of failure.
 */
extern ssize_t feenServiceWriteRaw (
    feenRC_t rc,
    uint8_t addr,
    const uint8_t * data,
    size_t length,
    char ** const message
    );

/** Convenience UART function.
 * @deprecated Not yet implemented.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] addr
 *      Address on the RS485 bus. Valid values range from
 *      @c 0 to @c 15.
 *      Special exception: If @a addr is @c 0x
 * @param [out] data Read data will be stored here.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return Number of bytes actuall read, or @c -1 in case of failure.
 */
extern ssize_t feenServiceReadRaw (
    feenRC_t rc,
    uint8_t addr,
    uint8_t * data,
    size_t length,
    char ** const message
    );

/** Talk to the Service Processor.
 *
 * Use this command to talk to the service processor if you trust the
 * rootcard to handle the protocol. In principle, you could use raw UARTs to talk
 * to the service processor, but this command is more efficient, and you don't have to
 * worry about parity and baud rate issues.
 *
 * Note that there is no PID input parameter. The protocol 
 * automatically increments the PID of the RS485 packet by one
 * for seach nodecard and seach command that is successfully 
 * sent.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] addr 
 *      Address on the RS485 bus.
 *      Valid Addresses range from @c 0 to @c 15.
 * @param [in] cmd
 *      Command of the Rootcard/SP (service processor) protocol.
 *      See the service processor spec for details.
 * @param [in] dataIn
 *      Auxilliary data of the command to be written.
 *      Valid data depends on @a cmd. 
 *      See the service processor spec for details.
 * @param [in] lengthIn
 *      length of @a dataIn
 * @param [out] pidOut
 *      PID of the service processors' answer
 * @param [out] cmdOut
 *      CMD of the service processors' answer.
 *      See the service processor spec for details.
 * @param [out] statOut
 *      STAT of the service processors' answer.
 *      See the service processor spec for details.
 * @param [out] dataOut
 *      Auxilliary data of the service processors' answer.
 *      The user has to allocate enough memory at the 
 *      location where @a dataOut points to such that all data
 *      can be stored there. Each service processor can return
 *      a maximum of 265 bytes.
 * @param [out] lengthOut
 *      Length of the auxilliary data was received.
 * @param [out] checksumOut
 *      CHK (checksum) part of the service processors' answer.
 * @param [out] message
 *      See @a feenIICWrite.
 * @return @c  0   in case of success.
 * @return @c -1   in case of unspecific error
 * @return @c -2   in case of service processor timeout
 * @return @c -3   in case of checksum error (the packet is returned nevertheless)
 * @return @c -4   in case of PID mismatch
 * @return @c -5   in case of CMD mismatch
 * @return @c -6   in case that STAT is non-zero.
 * @return @c -7   in case the service processor is BUSY
 */
extern int feenServiceCommand (
    feenRC_t rc,
    uint8_t addr,
    uint8_t cmd,
    const uint8_t * dataIn,
    uint8_t lengthIn,
    uint8_t * pidOut,
    uint8_t * cmdOut,
    uint8_t * statOut,
    uint8_t * dataOut,
    uint8_t * lengthOut,
    uint8_t * checksumOut,
    char ** const message
    );

/** Service processor convenience command to set the PID.
 * @deprecated Not yet implemented.
 *
 * This command sets the PID part of the RS485 packet
 * for the next command sent from the rootcard to
 * one of the service processors.
 * @param [in] rc See @a feenIICWrite.
 * @param [in] addr 
 *      Address on the RS485 bus.
 *      Valid Addresses range from @c 0 to @c 15.
 * @param [in] pid
 *      The value the next PID should have.
 * @param [out] message See @a feenIICWrite.
 * @return @c  0 in case of success
 * @return @c -1 in case of failure
 */
extern int feenServiceSetPid (
    feenRC_t rc,
    uint8_t addr,
    uint8_t pid,
    char ** const message
    );

/** Service processor convenice command to get the PID.
 * @deprecated Not yet implemented.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] addr
 *      Address on the RS485 bus.
 *      Valid Addresses range from @c 0 to @c 15.
 * @param [out] pid
 *      The current PID. The next packet sent on 
 *      RS485 will have that PID.
 * @param [out] message See @a feenIICWrite.
 * @return @c  0 in case of success
 * @return @c -1 in case of failure
 */
extern int feenServiceGetPid (
    feenRC_t rc,
    uint8_t addr,
    uint8_t * pid,
    char ** const message
    );

/** Service processor convenience command to reset all PIDs.
 * @deprecated Not yet implemented.
 *
 * The PIDs for all service processors are set to @c 0.
 * @param [in] rc See @a feenIICWrite.
 * @param [out] message See @a feenIICWrite.
 * @return @c  0 in case of success
 * @return @c -1 in case of failure
 */
extern int feenServiceResetPids (
    feenRC_t rc,
    char ** const message
    );

/** Issue an UPD read on the rootcard.
 * @deprecated Not yet implemented.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] addr 
 *      The IP address as a 32-bit integer
 *      in <b>host byte order</b>.
 * @param [in] port
 *      The IP port number as 16-bit integer
 *      in <b>host byte order</b>
 * @param [out] data Read data is stored here.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return Number of bytes actually read, or @c -1 in case of failure.
 */
extern ssize_t feenUDPRead (
    feenRC_t rc,
    uint32_t addr,
    uint16_t port,
    uint8_t * data,
    size_t length,
    char ** const message
    );

/** Issue an UDP write on the rootcard.
 * @deprecated Not yet implemented.
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [in] addr
 *      The IP address as a 32-bit integer
 *      in <b>host byte order</b>.
 * @param [in] port
 *      The IP port number as 16-bit integer
 *      in <b>host byte order</b>
 * @param [in] data Data to be written.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return Number of bytes actually written, or @c -1 in case of failure.
 */
extern ssize_t feenUDPWrite (
    feenRC_t rc,
    uint32_t addr,
    uint16_t port,
    const uint8_t * data,
    size_t length,
    char ** const message
    );

/** Transfer a file to the rootcard.
 *
 * This routine transfers a file from the frontend to the rootcard.
 * It does not use TFTP or FTP, but uses a custom protocol,
 * designed to be reliable.
 * @param [in] rc See @a feenIICWrite.
 * @param [in] filename
 *      Filename on the rootcard.
 *      Note that only paths under @c /var/ can be written.
 * @param [in] append
 *      If @a append is @c 0, the file is truncated if it
 *      already exists. If @a append is @c 1, the new data
 *      will be appended to the file.
 * @param [in] data Data to be transferred.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return Number of bytes actually transferred, or @c -1 in case of failure.
 */
extern ssize_t feenPushFile (
    feenRC_t rc,
    const char * filename,
    int append,
    const uint8_t * data,
    size_t length,
    char ** const message
    );

/** Create directory on rootcard.
 *
 * @param [in] rc rootcard
 * @param [in] filename Name of the directory to create
 * @param [out] message See @a feenIICWrite.
 * @return @c 0 in case of success, @c -1 otherwise.
 */
extern int feenMkDir(
    feenRC_t rc,
    const char * filename,
    char ** const message
    );

/** Transfer a file from the rootcard.
 *
 * This routine transfers a file from the rootcard to the frontend.
 * It uses the same protocol as @a feenPushFile.
 * @param [in] rc See @a feenIICWrite.
 * @param [in] filename Filename on the rootcard.
 * @param [out] data Transferred data will be stored here.
 * @param [in] length Length of @a data.
 * @param [out] message See @a feenIICWrite.
 * @return Bytes actually transferred, or @c -1 in case of failure.
 */
extern ssize_t feenPullFile (
    feenRC_t rc,
    const char * filename,
    uint8_t * data,
    size_t length,
    char ** const message
    );

/** Executes any command on the rootcard.
 * @deprecated Not yet implemented.
 *
 * This call tries to execute a command on the rootcard.
 * @a feenAnyCommand then tries to connect the standard input
 * of the program on the rootcard with the file descriptor @a infile
 * on the frontend, and tries to do all read/write operations 
 * in non-blocking mode if possible.
 * The standard output will be connected to @a outfile and
 * the standard error to @a errfile.
 * 
 * If a filedescriptor is invalid, i.e., if it is @c -1, then
 * no redirection for this filedescriptor will take place.
 *
 * However, the function itself blocks, just like all other functions
 * of the <b>feenlib</b> library. Execution of the command on the 
 * rootcard happens without a timeout. It is the user's responsibility
 * to kill the process on the rootcard.
 *
 * Example:
 * <pre>
 *  char * xargv[] = {"sh", "-c", "ls", "-l", "/var/", NULL};
 *  char * xenv[] = {NULL};
 *  int fdin, fdout, fderr;
 *  feenAnyCommand (
 *      fd, "/bin/sh",
 *      xargv, xenv,
 *      &fdin, &fdout, &fderr,
 *      NULL
 *      );
 *  dup2(fdin, STDIN_FILENO);
 *  dup2(fdout, STDOUT_FILENO);
 *  dup2(dferr, STDERR_FILENO);
 * </pre>
 */
extern int feenAnyCommand (
    feenRC_t rc,
    const char* path,
    char ** argv,
    char ** env,
    int* infile,
    int* outfile,
    int* errfile,
    char ** const message
    );

/** Get the status of the rootcard.
 *
 * This routine returns the status of the rootcard
 * in <b>host byte order</b>.
 *
 * Possible return values of @a status are:
 * <table>
 *      <tr>
 *          <th>name</th>
 *          <th>value</th>
 *          <th>meaning</th>
 *      </tr>
 *      <tr>
 *          <td>FEEN_SCANNING</td>
 *          <td>0x88000001</td>
 *          <td>    The transfer of a packet started,
 *                  but is not yet complete. The server process
 *                  on the rootcard is in the process of scanning for a new
 *                  packet.
 *          </td>
 *      </tr>
 *      <tr>
 *          <td>FEEN_INTER_WIP</td>
 *          <td>0x88000002</td>
 *          <td>    An internal process of the server process on the
 *                  rootcard is currently running.
 *          </td>
 *      </tr>
 *      <tr>
 *          <td>FEEN_EXTERN_WIP</td>
 *          <td>0x88000003</td>
 *          <td>    An external programm, spawned by the server process
 *                  on the rootcard is  currently running.
 *          </td>
 *      </tr>
 *      <tr>
 *          <td>FEEN_IDLE</td>
 *          <td>0x88000004</td>
 *          <td>    The server process on the rootcard is idle
 *                  and awaits a new command.
 *          </td>
 *      </tr>
 * </table>
 *
 * @param [in] rc See @a feenIICWrite.
 * @param [out] status Status of the rootcard.
 * @param [out] message See @a feenIICWrite.
 * @return @c -1 in case of failure, otherwise return the 
 *      process id (PID) of the process started on the rootcard.
 */
extern int feenStatus (
    feenRC_t rc,
    uint32_t * status,
    char ** const message
    );

/** Restart the server in the rootcard.
 *
 * This routine resets the listening server
 * on the rootcard. It will be restarted automatically.
 * All running processes spawned by the server on the
 * rootcard are killed.
 * @param [in] rc See @a feenIICWrite.
 * @param [out] message See @a feenIICWrite.
 * @return @c  0 in case of success.
 * @return @c -1 in case of failure.
 */
extern int feenRestart (
    feenRC_t rc,
    char ** const message
    );

/** Initialize rootcard communication. 
 *
 * This function initializes a rootcard variable.
 * 
 * @param [out] rc
 *      Variable that holds the communication data for 
 *      a given rootcard.
 * @param [in] addr
 *      IP address of the rootcard to talk to in 
 *      host byte order.
 * @param [in] dataport 
 *      IP port number in host byte order of the underlying 
 *      protocol. Use @a FEEN_DEFAULT_DATAPORT if you are unsure.
 * @param [in] controlport 
 *      IP port number in host byte order of the underlying
 *      protocol. Use @a FEEN_DEFAULT_CONTROLPORT if you are unsure.
 * @param [out] message See @a feenIICWrite.
 * @return  0 in case of success.
 * @return -1 in case of failure.
 */
extern int feenInitRootcardRaw (
    feenRC_t * rc,
    uint32_t addr,
    uint16_t dataport,
    uint16_t controlport,
    char ** const message
    );

/** Initialize rootcard communication.
 *
 * See also @a feenInitRootcardRaw
 *
 * @param [out] rc See @a feenInitRootcardRaw
 * @param [in] addr
 *      IP address of the rootcard as a C-string,
 *      representing a numeric IP, e.g.,
 *      "132.199.10.10"
 * @param [in] dataport 
 *      IP port number in host byte order of the underlying 
 *      protocol. Use @a FEEN_DEFAULT_DATAPORT if you are unsure.
 * @param [in] controlport 
 *      IP port number in host byte order of the underlying
 *      protocol. Use @a FEEN_DEFAULT_CONTROLPORT if you are unsure.
 * @param [out] message See @a feenIICWrite.
 * @return  0 in case of success.
 * @return -1 in case of failure.
 */
extern int feenInitRootcardNumeric (
    feenRC_t * rc,
    const char * addr,
    uint16_t dataport,
    uint16_t controlport,
    char ** const message
    );

/** Initialize rootcard communication.
 *
 * See also @a feenInitRootcardRaw.
 *
 * @param [out] rc See @a feenInitRootcardRaw.
 * @param [in] addr
 *      IP address of the rootcard as a C-string,
 *      representing a named IP address, e.g.,
 *      "rootcard01.qpace.example.com"
 * @param [in] dataport 
 *      IP port number in host byte order of the underlying 
 *      protocol. Use @a FEEN_DEFAULT_DATAPORT if you are unsure.
 * @param [in] controlport 
 *      IP port number in host byte order of the underlying
 *      protocol. Use @a FEEN_DEFAULT_CONTROLPORT if you are unsure.
 * @param [out] message See @a feenIICWrite.
 * @return  0 in case of success.
 * @return -1 in case of failure.
 */
extern int feenInitRootcardNamed (
    feenRC_t * rc,
    const char * addr,
    uint16_t dataport,
    uint16_t controlport,
    char ** const message
    );

/** Initialize rootcard communication.
 *
 * This is a wrapper to @a feenInitRootcardNumeric and
 * @a feenInitRootcardNamed. It uses simple heuristics to distinguish
 * between a named IP (like rc01.qpace.whatever) and numeric IPs
 * (like 10.4.0.1).
 * If the parameter @a addr consists only of dots and decimals,
 * then a numeric IP is assumed, otherwise a named IP is assumed.
 *
 * For a description of the parameters, see
 * @a feenInitRootcardNamed.
 */
extern int feenInitRootcardAuto (
    feenRC_t * rc,
    const char * addr,
    uint16_t dataport,
    uint16_t controlport,
    char ** const message
    );

/** Fully automated rootcard initialization.
 *
 * Beware: This function contains a hard-coded mapping beween
 * rootcard numeric IPs and contol/data port numbers.
 */
extern int feenInitRootcardFullAuto (
    feenRC_t * rc,
    const char * addr,
    char ** const message
    );

/** Set the timeout for a rootcard.
 * This function sets the timeout (in seconds) for a given
 * rootcard. Most feenXXX functions will respect this
 * timeout, and return an error in case the function timed out.
 * @param [in,out] rc rootcard
 * @param [in] timeout timeout in seconds
 */
extern void feenRootcardSetTimeout(
    feenRC_t * rc,
    int timeout
    );

/** Get the timeout for a rootcard.
 * 
 * Note: The default timeout for a rootcard
 * is 60 seconds.
 *
 * @param [in] rc rootcard
 * @return timeout in seconds
 */
extern int feenRootcardGetTimeout(
    feenRC_t rc
    );

/** Set the verbosity of RS485 debug messages on the console.
 *
 * @param [in] rc rootcard
 * @param [in] level Verbosity Level. Valid range is 0 thru 3.
 * @param [out] message See @a feenIICWrite.
 * @return @c 0 in case of success, @c -1 otherwise.
 */
extern int feenSetRS485Verbosity(
    feenRC_t rc,
    int level,
    char ** const message
    );

/** Returns the SVN Id of the header file.
 * 
 * This function returns the SVN (subversion)
 * identity of the header file "feenlib.h".
 * Note that the SVN Id is <b>not</b> the currrent
 * head revision of the SVN repository.
 * It is that revsion, in which the file in question
 * (feenlib.h) was last changed.
 */
extern const char* feenSVNIdentityOfFeenlibH(void);

/** Returns the SVN Id of the library file.
 * 
 * This function returns the SVN (subversion)
 * identity of the library file "feenlib.c".
 * Note that the SVN Id is <b>not</b> the currrent
 * head revision of the SVN repository.
 * It is that revsion, in which the file in question
 * (feenlib.d) was last changed.
 */
extern const char* feenSVNIdentityOfFeenlibC(void);

/** Returns the HG Id of the library.
 *
 * This function returns the HG (mercurial)
 * identity of the whole library. Note that
 * this is the identity of the current head,
 * in contrast to the SVN identity.
 * This is done through the makefile, and not
 * part of mercurial.
 */
extern const char* feenHGIdentity(void);

/** Returns the protocol version running on the rootcard.
 *
 * @param [in] rc rootcard
 * @param [out] message See @a feenIICWrite.
 * @return @c -1 in case of failure, otherwise the
 *      protocol version is returned.
 */
extern int feenProtocolVersion(
    feenRC_t rc,
    char ** const message
    );

/** Close rootcard.
 * This routine closes all open connections of a given rootcard.
 * After this call, the parameter @a rc needs to be
 * initialized again to be usable.
 * @param [in] rc rootcard
 * @param [out] message See @a feenIICWrite.
 * @return @c 0 in case of success, @c -1 otherwise
 */
extern int feenRootCardClose(
    feenRC_t rc,
    char** const message);

/** Graceful Restart.
 * @param [in,out] rc rootcard
 * @param [out] message See @a feenIICWrite
 * @return @c 0 in case of success, @c -1 otherwise.
 */
extern int feenGraceRestart(
    feenRC_t* rc,
    char** const message
    );

/** Cooperative locking.
 * @deprecated Not yet implemented.
 *
 * @param [in] rc rootcard
 * @param [in] name Name of the variable
 * @param [in] reading Just read, or read and write?
 * @param [in] overwrite what to do if the variable already exists
 * @param [in,out] value
 * @param [in,out] ttl time-to-live
 * @param [out] message See @a feenIICWrite
 * @return @c 0 in case of success, @ -1 otherwise
 *
 * If @a reading is @c 1, then the value of the variable ist just
 * read, and returned in @a value. If the variable does not exist,
 * @a feenLock returns @c -1 and @a value is not changed.
 * One can also pass @c NULL for @a value. In that case,
 * @a feenLock just checks if the variable is present.
 * The @a ttl parameter (time-to-live) is also returned, if
 * a non-NULL pointer is passed.
 *
 * If @a reading is @c 0, and @a overwrite is @c 1,
 * then the function tries to read the
 * old value, and then set the new value. In that case,
 * @a value is a value-result parameter. First, it holds the 
 * value the variable should be set to, afterwards it holds the
 * old value of the variable.
 * The parameter @a ttl (time-to-live) sets a timeout of the
 * lock in seconds. Once the lock has expired, it is
 * released automatically.
 *
 * If @a reading is @c 0 and @a overwrite is also @c 0, then
 * the variable will not be overwritten. Thus,
 * @a feenLock will return @c -1 if the variable already exists.
 */
extern int feenLock(
    feenRC_t rc,
    const char* name,
    int reading,
    int overwrite,
    int32_t* value,
    int32_t* ttl,
    char** const message);

/** Release cooperative locks.
 * @deprecated Not yet implemented.
 *
 * @param [in] rc rootcard
 * @param [in] name name of the variable
 * @param [out] message See @a feenIICWrite.
 * Release a cooperative lock previously set with
 * @a feenLock.
 */
extern int feenLockRelease(
    feenRC_t rc,
    const char* name,
    char** const message);

/** Set time time on the rootcard.
 *
 * @param [in] rc rootcard
 * @param [in] year Year (e.g. 2008)
 * @param [in] month Month (1-12)
 * @param [in] day Day (1-31)
 * @param [in] hour Hour (0-23)
 * @param [in] minute Minute (0-59)
 * @param [in] second Second (0-59)
 * @param [out] message See @a feenIICWrite.
 * @return @c 0 in case of success, @c -1 otherwise
 */
extern int feenSetTime(
    feenRC_t rc,
    int year, int month, int day,
    int hour, int minute, int second,
    char** const message);

/** Get time time on the rootcard.
 *
 * @param [in] rc rootcard
 * @param [in] year Year (e.g. 2008)
 * @param [in] month Month (1-12)
 * @param [in] day Day (1-31)
 * @param [in] hour Hour (0-23)
 * @param [in] minute Minute (0-59)
 * @param [in] second Second (0-59)
 * @param [out] message Seen @c feenIICWrite
 * @return @c 0 in case of success, @c -1 otherwise
 */
extern int feenGetTime(
    feenRC_t rc,
    int* year, int* month, int* day,
    int* hour, int* minute, int* second,
    char** const message);

/** Write to TCP server.
 * @deprecated Not yet implemented.
 * @param [in] rc rootcard
 * @param [in] server Server IP
 * @param [in] port Server TCP port
 * @param [in] data data to be written
 * @param [in] length size of data to be written
 * @param [out] message See @a feenIICWrite
 * @return Number of bytes actually written, or @c -1 in case of failure
 */
extern ssize_t feenTCPWrite(
    feenRC_t rc,
    const char* server,
    uint16_t port,
    const uint8_t* data,
    size_t length,
    char** const message);

/** Read from TCP server.
 * @deprecated Not yet implemented.
 * @param [in] rc rootcard
 * @param [in] server Server IP
 * @param [in] port Server TCP port
 * @param [out] data read data
 * @param [in] length of data to be read
 * @param [out] message See @a feenIICWrite
 * @return Number of bytes actually read, or @c -1 in case of failure
 */
extern ssize_t feenTCPRead(
    feenRC_t rc,
    const char* server,
    uint16_t port,
    uint8_t* data,
    size_t length,
    char** const message);

/** Get stats of a file on the rootcard.
 *
 * @param [in] rc rootcard
 * @param [in] filename filename
 * @param [out] exists This returns @c 1 if the file exists,
 *      and @c 0 otherwise
 * @param [out] size file size
 * @param [out] atime Returns the access time
 * @param [out] mtime Returns the modification time
 * @param [out] ctime Returns the status change time
 * @param [out] message See @a feenIICWrite
 * @return @c 0 in case of success, @c -1 otherwise
 */
extern int feenFileStat(
    feenRC_t rc,
    const char* filename,
    int* exists,
    size_t* size,
    time_t* atime,
    time_t* mtime,
    time_t* ctime,
    char** const message);

/** Unlink (delete) file on the rootcard.
 *
 * @param [in] rc rootcard
 * @param [in] filename filename
 * @param [out] message See @a feenIICWrite
 * @return @c 0 in case of success, @c -1 otherwise.
 */
extern int feenFileUnlink(
    feenRC_t rc,
    const char* filename,
    char** const message);

/** Reconnect a rootcard.
 * @param [in] rc pointer to rootcard
 * @param [out] message See @a feenIICWrite
 * @return @c 0 in case of success, @c -1 otherwise
 */
extern int feenReconnect(
    feenRC_t* rc,
    char** const message);

#endif /* __uClinux__ */

#endif /* FEENLIB_H_ */

